@@ -8,12 +8,12 @@
        (http://davis.wpi.edu/~matt/courses/clipping/).
 \author José Luis García Pallero, jgpallero@gmail.com
 \date 14 de mayo de 2011
-\section Licencia Licencia
-Copyright (c) 2011, José Luis García Pallero. All rights reserved.
-
+\copyright
+Copyright (c) 2011-2020, José Luis García Pallero. All rights reserved.
+\par
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:
-
+\par
 - Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright notice, this
@@ -22,7 +22,7 @@ are permitted provided that the following conditions are met:
 - Neither the name of the copyright holders nor the names of its contributors
   may be used to endorse or promote products derived from this software without
   specific prior written permission.
-
+\par
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
@@ -830,7 +830,8 @@ int Paso1Greiner(vertPoliClip* poliBas,
                         //un extremo e intersecciones colineales donde se
                         //comparte más de un punto
                         if((intersec==GEOC_SEG_INTERSEC_EXTREMO_NO_COLIN)||
-                           (intersec==GEOC_SEG_INTERSEC_EXTREMO_COLIN))
+                           (intersec==GEOC_SEG_INTERSEC_EXTREMOS_NO_COLIN)||
+                           (intersec==GEOC_SEG_INTERSEC_EXTREMOS_COLIN))
                         {
                             if((xI==auxC->xP)&&(yI==auxC->yP))
                             {
@@ -1230,17 +1231,26 @@ polig* PoliBoolGreiner(vertPoliClip* poliBas,
                        vertPoliClip* poliRec,
                        const enum GEOC_OP_BOOL_POLIG op,
                        const double facPer,
+                       const int compruebaHuecosUnion,
                        size_t* nIntersec,
                        size_t* nPerturb)
 {
+    //índice para recorrer bucles
+    size_t i=0;
     //factor de perturbación
     double factor=0.0;
     //variables de posición
     int posBas=0,posRec=0;
     //identificador de error
     int idError=GEOC_ERR_NO_ERROR;
-    //polígono auxiliar
+    //variables auxiliares
+    int enPol1=GEOC_PTO_FUERA_POLIG,enPol2=GEOC_PTO_FUERA_POLIG;
+    size_t posIni=0;
+    double xAux=0.0,yAux=0.0;
+    //polígonos auxiliares
     polig* ba=NULL;
+    polig* aux1=NULL;
+    polig* aux2=NULL;
     //polígono de salida
     polig* resultado=NULL;
     ////////////////////////////////////////////////////////////////////////////
@@ -1453,6 +1463,76 @@ polig* PoliBoolGreiner(vertPoliClip* poliBas,
     }
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
+    //inicializo los atributos a 1
+    for(i=0;i<resultado->nPolig;i++)
+    {
+        resultado->atr[i] = 1;
+    }
+    //compruebo si hemos trabajado con unión
+    if(compruebaHuecosUnion&&(op==GeocOpBoolUnion))
+    {
+        //convierto los polígonos de trabajo en estructuras polig
+        aux1 = CreaPoligPoliClip(poliBas,1);
+        aux2 = CreaPoligPoliClip(poliRec,1);
+        //comprobamos los posibles errores
+        if((aux1==NULL)||(aux2==NULL))
+        {
+            //liberamos la memoria asignada
+            LibMemPolig(resultado);
+            LibMemPolig(aux1);
+            LibMemPolig(aux2);
+            //mensaje de error
+            GEOC_ERROR("Error de asignación de memoria en la llamada a "
+                       "'CreaPoligPoliClip' para la comprobación de huecos en "
+                       "la operación OR");
+            //salimos de la función
+            return NULL;
+        }
+        //recorro todos los polígonos generados en la unión
+        for(i=0;i<resultado->nPolig;i++)
+        {
+            //posición de inicio del polígono
+            posIni = resultado->posIni[i];
+            //genero un punto dentro del polígono
+            GeneraPtoEnPoligono(&(resultado->x[posIni]),&(resultado->y[posIni]),
+                                resultado->nVert[i],1,1,0.0,1,&xAux,&yAux);
+            //comprobamos las coordenadas de salida
+            if (EsGeocNan(xAux)||EsGeocNan(yAux))
+            {
+                //digo que es un hueco
+                resultado->atr[i] = 0;
+            }
+            else
+            {
+                //compruebo si el punto está dentro de alguno de los dos
+                //polígonos de trabajo
+                posIni = aux1->posIni[0];
+                enPol1 = PtoEnPoligonoVerticeBordeDouble(xAux,yAux,
+                                                         &(aux1->x[posIni]),
+                                                         &(aux1->y[posIni]),
+                                                         aux1->nVert[0],1,1,
+                                                         0.0,1);
+                posIni = aux2->posIni[0];
+                enPol2 = PtoEnPoligonoVerticeBordeDouble(xAux,yAux,
+                                                         &(aux2->x[posIni]),
+                                                         &(aux2->y[posIni]),
+                                                         aux2->nVert[0],1,1,
+                                                         0.0,1);
+                //si el polígono no pertenece a ninguno de los dos originales es
+                //un hueco
+                if((!enPol1)&&(!enPol2))
+                {
+                    //lo marco como hueco
+                    resultado->atr[i] = 0;
+                }
+            }
+        }
+        //liberamos la memoria asignada
+        LibMemPolig(aux1);
+        LibMemPolig(aux2);
+    }
+    ////////////////////////////////////////////////////////////////////////////
+    ////////////////////////////////////////////////////////////////////////////
     //salimos de la función
     return resultado;
 }
@@ -1462,6 +1542,7 @@ polig* PoliBoolGreinerMult(const polig* poliBas,
                            const polig* poliRec,
                            const enum GEOC_OP_BOOL_POLIG op,
                            const double facPer,
+                           const int compruebaHuecosUnion,
                            size_t* nIntersec,
                            size_t* nPerturb)
 {
@@ -1527,10 +1608,10 @@ polig* PoliBoolGreinerMult(const polig* poliBas,
             {
                 //comprobamos si los restángulos que encierran a los polígonos
                 //son disjuntos o no
-                pr = RectDisjuntos(poliBas->xMin[i],poliBas->xMax[i],
-                                   poliBas->yMin[i],poliBas->yMax[i],
-                                   poliRec->xMin[j],poliRec->xMax[j],
-                                   poliRec->yMin[j],poliRec->yMax[j]);
+                pr = GEOC_RECT_DISJUNTOS(poliBas->xMin[i],poliBas->xMax[i],
+                                         poliBas->yMin[i],poliBas->yMax[i],
+                                         poliRec->xMin[j],poliRec->xMax[j],
+                                         poliRec->yMin[j],poliRec->yMax[j]);
                 //comprobamos los casos particulares si los rectángulos son
                 //disjuntos
                 if(pr&&(op==GeocOpBoolInter))
@@ -1631,7 +1712,8 @@ polig* PoliBoolGreinerMult(const polig* poliBas,
                 return NULL;
             }
             //recortamos
-            poligAux = PoliBoolGreiner(poligBas,poligRec,op,facPer,&nInt,&nPer);
+            poligAux = PoliBoolGreiner(poligBas,poligRec,op,facPer,
+                                       compruebaHuecosUnion,&nInt,&nPer);
             //comprobamos los posibles errores
             if(poligAux==NULL)
             {
