@@ -7,12 +7,12 @@
 \author José Luis García Pallero, jgpallero@gmail.com
 \note Este fichero contiene funciones paralelizadas con OpenMP.
 \date 20 de abril de 2011
-\section Licencia Licencia
-Copyright (c) 2011, José Luis García Pallero. All rights reserved.
-
+\copyright
+Copyright (c) 2011-2020, José Luis García Pallero. All rights reserved.
+\par
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:
-
+\par
 - Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright notice, this
@@ -21,7 +21,7 @@ are permitted provided that the following conditions are met:
 - Neither the name of the copyright holders nor the names of its contributors
   may be used to endorse or promote products derived from this software without
   specific prior written permission.
-
+\par
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
@@ -93,6 +93,7 @@ polig* IniciaPoligVacio(void)
     sal->yMin = NULL;
     sal->yMax = NULL;
     sal->area = NULL;
+    sal->atr = NULL;
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
     //salimos de la función
@@ -342,6 +343,8 @@ polig* CreaPolig(const double* x,
                  const size_t incY,
                  int* idError)
 {
+    //variable para recorrer bucles
+    size_t i=0;
     //número de puntos
     size_t ptos=0;
     //número máximo de elementos de los vectores de coordenadas
@@ -383,7 +386,7 @@ polig* CreaPolig(const double* x,
     //paralelización con OpenMP
 #if defined(_OPENMP)
 #pragma omp parallel sections default(none) \
- shared(posNanX,x,nNanX,posNanY,y,nNanY)
+ shared(posNanX,nElem,incX,x,nNanX,posNanY,incY,y,nNanY)
 #endif
 {
 #if defined(_OPENMP)
@@ -548,6 +551,29 @@ polig* CreaPolig(const double* x,
     }
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
+    //inicializamos el vector de atributos
+    sal->atr = (int*)malloc(sal->nPolig*sizeof(int));
+    //comprobamos los posibles errores
+    if(sal->atr==NULL)
+    {
+        //liberamos la memoria asignada
+        LibMemPolig(sal);
+        free(posNanX);
+        free(posNanY);
+        //asignamos la variable de error
+        *idError = GEOC_ERR_ASIG_MEMORIA;
+        //mensaje de error
+        GEOC_ERROR("Error de asignación de memoria");
+        //salimos de la función
+        return NULL;
+    }
+    //Inicializamos el vector de atributos a 0
+    for(i=0;i<sal->nPolig;i++)
+    {
+        sal->atr[i] = 0;
+    }
+    ////////////////////////////////////////////////////////////////////////////
+    ////////////////////////////////////////////////////////////////////////////
     //liberamos la memoria asignada
     free(posNanX);
     free(posNanY);
@@ -569,11 +595,12 @@ void EnlazaCamposPolig(polig* poliEnt,
         free(poliSal->x);
         free(poliSal->y);
     }
-    //comprobamos si hay algún polígono en los vectores de posiciones
+    //compruebo si hay algún polígono en los vectores de posiciones y atributos
     if(poliSal->nPolig)
     {
         free(poliSal->posIni);
         free(poliSal->nVert);
+        free(poliSal->atr);
     }
     //comprobamos si hay límites calculados
     if(poliSal->hayLim)
@@ -604,6 +631,7 @@ void EnlazaCamposPolig(polig* poliEnt,
     poliSal->yMax = poliEnt->yMax;
     poliSal->hayArea = poliEnt->hayArea;
     poliSal->area = poliEnt->area;
+    poliSal->atr = poliEnt->atr;
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
     //salimos de la función
@@ -683,7 +711,7 @@ polig* CopiaPolig(const polig* poli,
         if(poli->hayArea)
         {
             //calculamos las áreas
-            *idError = CalcAreaPolig(sal);
+            *idError = CalcAreaPolig(sal,1.0,0,0.0,0.0);
             //comprobamos los posibles errores
             if((*idError)!=GEOC_ERR_NO_ERROR)
             {
@@ -741,6 +769,7 @@ int AnyadePoligPolig(polig* poli,
     poli->y = (double*)realloc(poli->y,nElem*sizeof(double));
     poli->posIni = (size_t*)realloc(poli->posIni,nPolig*sizeof(size_t));
     poli->nVert = (size_t*)realloc(poli->nVert,nPolig*sizeof(size_t));
+    poli->atr = (int*)realloc(poli->atr,nPolig*sizeof(int));
     //reasignamos también para los posibles vectores de límites y superficies
     if(poli->hayLim)
     {
@@ -755,7 +784,7 @@ int AnyadePoligPolig(polig* poli,
     }
     //comprobamos los posibles errores en las asignaciones obligatorias
     if((poli->x==NULL)||(poli->y==NULL)||(poli->posIni==NULL)||
-       (poli->nVert==NULL))
+       (poli->nVert==NULL)||(poli->atr==NULL))
     {
         //mensaje de error
         GEOC_ERROR("Error de asignación de memoria");
@@ -810,6 +839,7 @@ int AnyadePoligPolig(polig* poli,
         //copiamos las posiciones de inicio actualizadas y el número de vértices
         poli->posIni[poli->nPolig+i] = anyade->posIni[i]+pos;
         poli->nVert[poli->nPolig+i] = anyade->nVert[i];
+        poli->atr[poli->nPolig+i] = anyade->atr[i];
     }
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
@@ -997,6 +1027,11 @@ void LibMemPolig(polig* datos)
         {
             free(datos->area);
         }
+        //liberamos la memoria asignada al vector de atributos
+        if(datos->atr!=NULL)
+        {
+            free(datos->atr);
+        }
         //liberamos la memoria asignada a la estructura
         free(datos);
     }
@@ -1088,7 +1123,7 @@ void LimitesPoligono(const double* x,
     //paralelización con OpenMP
 #if defined(_OPENMP)
 #pragma omp parallel sections default(none) \
- shared(x,posXMin,posXMax,y,posYMin,posYMax)
+ shared(incX,x,nElem,posXMin,posXMax,incY,y,posYMin,posYMax)
 #endif
 {
     //posiciones en el vector X
@@ -1136,7 +1171,7 @@ void LimitesPoligonosPolig(const double* x,
     //paralelización con OpenMP
 #if defined(_OPENMP)
 #pragma omp parallel for default(none) schedule(dynamic) \
- shared(posIni,x,y,nVert,xMin,xMax,yMin,yMax) \
+ shared(nPolig,posIni,restaPosIni,incX,x,incY,y,nVert,xMin,xMax,yMin,yMax) \
  private(i,pI)
 #endif
     //recorremos el número de polígonos
@@ -1155,38 +1190,121 @@ void LimitesPoligonosPolig(const double* x,
 }
 /******************************************************************************/
 /******************************************************************************/
-int CalcAreaPolig(polig* poli)
+int CalcAreaPolig(polig* poli,
+                  const double facCoor,
+                  const int geo,
+                  const double a,
+                  const double f)
 {
+    //índices para recorrer bucles
+    size_t i=0,j=0;
+    //variable de posición
+    size_t pos=0;
+    //longitud origen
+    double lon0=0.0;
+    //coordenadas de trabajo
+    double* x=NULL;
+    double* y=NULL;
     //variable de estado (salida)
     int estado=GEOC_ERR_NO_ERROR;
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
     //comprobamos salida rápida
-    if((poli->nPolig==0)||(poli->hayArea))
+    if(poli->nPolig==0)
     {
-        //salimos de la función si la estructura no contiene polígonos o si
-        //éstos ya tienen calculada su superficie
+        //salimos de la función si la estructura
         return estado;
     }
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
-    //asignamos memoria para el vector de superficies
-    poli->area = (double*)malloc((poli->nPolig)*sizeof(double));
-    //comprobamos los posibles errores
-    if(poli->area==NULL)
+    //comprobamos si la estructura no contiene área calculada
+    if(!(poli->hayArea))
     {
-        //mensaje de error
-        GEOC_ERROR("Error de asignación de memoria");
-        //salimos de la función
-        return GEOC_ERR_ASIG_MEMORIA;
+        //asignamos memoria para el vector de superficies
+        poli->area = (double*)malloc((poli->nPolig)*sizeof(double));
+        //comprobamos los posibles errores
+        if(poli->area==NULL)
+        {
+            //mensaje de error
+            GEOC_ERROR("Error de asignación de memoria");
+            //salimos de la función
+            return GEOC_ERR_ASIG_MEMORIA;
+        }
+        //indicamos que sí hay superficies
+        poli->hayArea = 1;
+    }
+    ////////////////////////////////////////////////////////////////////////////
+    ////////////////////////////////////////////////////////////////////////////
+    //comprobamos si trabajamos sobre la superficie del elipsoide
+    if(geo||(facCoor!=1.0))
+    {
+        //asignamos memoria para los vectores auxiliares
+        x = (double*)malloc((poli->nElem)*sizeof(double));
+        y = (double*)malloc((poli->nElem)*sizeof(double));
+        //comprobamos posibles errores
+        if((x==NULL)||(y==NULL))
+        {
+            //volvemos a indicar que no hay áreas
+            poli->hayArea = 0;
+            //liberamos la memoria asignada
+            free(poli->area);
+            free(x);
+            free(y);
+            //mensaje de error
+            GEOC_ERROR("Error de asignación de memoria");
+            //salimos de la función
+            return GEOC_ERR_ASIG_MEMORIA;
+        }
+        //copiamos las coordenadas y aplicamos factor de escala
+        //si estamos sobre el elipsoide, este paso se hace luego
+        if(!geo)
+        {
+            for(i=0;i<poli->nElem;i++)
+            {
+                x[i] = poli->x[i]*facCoor;
+                y[i] = poli->y[i]*facCoor;
+            }
+        }
+        else
+        {
+            //recorremos los polígonos
+            for(i=0;i<poli->nPolig;i++)
+            {
+                //posición de inicio en los vectores de coordenadas
+                pos = poli->posIni[i];
+                //calculamos la longitud media del polígono de trabajo
+                lon0 = Media(&(poli->x[pos]),poli->nVert[i],1);
+                //recorremos los vértices del polígono
+                for(j=0;j<poli->nVert[i];j++)
+                {
+                    //proyectamos, aplicando el factor de escala
+                    ProjCilinEquivLambertLat0Ec(poli->y[pos+j]*facCoor,
+                                                poli->x[pos+j]*facCoor,
+                                                lon0*facCoor,a,f,&x[pos+j],
+                                                &y[pos+j]);
+                }
+            }
+        }
+    }
+    else
+    {
+        //no hace falta copiarlas a los vectores auxiliares, sólo enlazarlas
+        x = poli->x;
+        y = poli->y;
     }
-    //indicamos que sí hay superficies
-    poli->hayArea = 1;
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
     //calculamos las superficies de todos los polígonos
-    AreaPoligonosSimplesPolig(poli->x,poli->y,1,1,poli->posIni,poli->nVert,
-                              poli->nPolig,0,poli->area);
+    AreaPoligonosSimplesPolig(x,y,1,1,poli->posIni,poli->nVert,poli->nPolig,0,
+                              poli->area);
+    ////////////////////////////////////////////////////////////////////////////
+    ////////////////////////////////////////////////////////////////////////////
+    //liberamos la memoria utilizada
+    if(geo||(facCoor!=1.0))
+    {
+        free(x);
+        free(y);
+    }
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
     //salimos de la función
@@ -1208,7 +1326,7 @@ double AreaPoligonoSimple(const double* x,
     double area=0.0;
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
-    //si el primero punto es el mismo que el último, restamos una unidad al
+    //si el primer punto es el mismo que el último, restamos una unidad al
     //número de elementos pasado, ya que el algoritmo está preparado para
     //trabajar con un vector en el que no se repite el primer punto
     if((x[0]==x[(nElem-1)*incX])&&(y[0]==y[(nElem-1)*incY]))
@@ -1255,7 +1373,7 @@ void AreaPoligonosSimplesPolig(const double* x,
     //paralelización con OpenMP
 #if defined(_OPENMP)
 #pragma omp parallel for default(none) schedule(dynamic) \
- shared(posIni,area,x,y,nVert) \
+ shared(nPolig,posIni,restaPosIni,area,incX,x,incY,y,nVert) \
  private(i,pI)
 #endif
     //recorremos el número de polígonos
@@ -1285,7 +1403,7 @@ void EscalaYTrasladaPolig(polig* poli,
     //paralelización con OpenMP
 #if defined(_OPENMP)
 #pragma omp parallel sections default(none) \
- shared(poli)
+ shared(poli,escalaX,trasladaX,escalaY,trasladaY)
 #endif
 {
 #if defined(_OPENMP)
@@ -1338,7 +1456,7 @@ void TrasladaYEscalaPolig(polig* poli,
     //paralelización con OpenMP
 #if defined(_OPENMP)
 #pragma omp parallel sections default(none) \
- shared(poli)
+ shared(poli,escalaX,trasladaX,escalaY,trasladaY)
 #endif
 {
 #if defined(_OPENMP)
@@ -1410,20 +1528,34 @@ void MuevePolig(polig* poli,
 /******************************************************************************/
 /******************************************************************************/
 int AligeraPolig(polig* poli,
+                 const int esf,
+                 const double facCoor,
                  const double tol,
+                 const int paralelizaTol,
                  const enum GEOC_DPEUCKER_ROBUSTO robusto,
-                 const size_t nPtosRobusto,
-                 const size_t nSegRobusto)
+                 const size_t nSegRobOrig,
+                 const size_t nSegRobAuto,
+                 const double a,
+                 const double f)
 {
     //índices para recorrer bucles
     size_t i=0,j=0;
+    //tolerancia angular
+    double tolAng=fabs(GEOC_ARC_RES_ANG);
     //posición inicial del polígono de trabajo y número de puntos del aligerado
-    size_t posIni=0,nPtos=0;
+    size_t posIni=0,nVert=0,nPtos=0;
     //puntos colineales
     int colin=0;
     //coordenadas de los polígonos aligerados
     double* x=NULL;
     double* y=NULL;
+    double* xF=NULL;
+    double* yF=NULL;
+    //vértices del polígono de entrada más largo
+    size_t nVertMax=0;
+    //matriz de rotación y coordenadas en el sistema rotado
+    double mRot[3][3];
+    double latCR=0.0;
     //estructura auxiliar
     polig* aux=NULL;
     //vector de posiciones después del aligerado
@@ -1452,20 +1584,65 @@ int AligeraPolig(polig* poli,
     }
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
+    //comprobamos si hay factor de escala para las coordenadas
+    if(facCoor!=1.0)
+    {
+        //vértices del polígono más largo
+        nVertMax = MaximoSizeT(poli->nVert,poli->nPolig,1);
+        //asignamos memoria para las coordenadas con factor de escala
+        xF = (double*)malloc(nVertMax*sizeof(double));
+        yF = (double*)malloc(nVertMax*sizeof(double));
+        //comprobamos posibles errores
+        if((xF==NULL)||(yF==NULL))
+        {
+            //liberamos la memoria asignada
+            LibMemPolig(aux);
+            free(xF);
+            free(yF);
+            //mensaje de error
+            GEOC_ERROR("Error de asignación de memoria");
+            //salimos de la función
+            return GEOC_ERR_ASIG_MEMORIA;
+        }
+    }
+    ////////////////////////////////////////////////////////////////////////////
+    ////////////////////////////////////////////////////////////////////////////
     //recorremos el número de polígonos almacenados
     for(i=0;i<poli->nPolig;i++)
     {
         //extraemos la posición inicial del polígono de trabajo
         posIni = poli->posIni[i];
+        //número de vértices del polígono
+        nVert = poli->nVert[i];
+        //comprobamos si hay factor de escala para las coordenadas
+        if(facCoor!=1.0)
+        {
+            //copiamos las coordenadas y aplicamos el factor de escala
+            for(j=0;j<nVert;j++)
+            {
+                xF[j] = facCoor*(poli->x[posIni+j]);
+                yF[j] = facCoor*(poli->y[posIni+j]);
+            }
+        }
+        else
+        {
+            //las coordenadas del polígono se quedan como están
+            xF = &(poli->x[posIni]);
+            yF = &(poli->y[posIni]);
+        }
         //aligeramos el polígono de trabajo
-        pos = AligeraPolilinea(&(poli->x[posIni]),&(poli->y[posIni]),
-                               poli->nVert[i],1,1,tol,robusto,nPtosRobusto,
-                               nSegRobusto,&nPtos);
+        pos = AligeraPolilinea(xF,yF,nVert,1,1,tol,paralelizaTol,robusto,
+                               nSegRobOrig,nSegRobAuto,esf,&nPtos);
         //comprobamos posibles errores
         if(pos==NULL)
         {
             //liberamos la memoria asignada
             LibMemPolig(aux);
+            if(facCoor!=1.0)
+            {
+                free(xF);
+                free(yF);
+            }
             //mensaje de error
             GEOC_ERROR("Error de asignación de memoria");
             //salimos de la función
@@ -1474,13 +1651,35 @@ int AligeraPolig(polig* poli,
         //comprobamos si el polígono se ha quedado en tres puntos
         if(nPtos==3)
         {
-            //comprobamos si los tres puntos son colineales
-            colin = TresPuntosColineales2D(poli->x[posIni+pos[0]],
-                                           poli->y[posIni+pos[0]],
-                                           poli->x[posIni+pos[1]],
-                                           poli->y[posIni+pos[1]],
-                                           poli->x[posIni+pos[2]],
-                                           poli->y[posIni+pos[2]]);
+            //distinguimos entre trabajo sobre la esfera y sobre el plano
+            if(esf)
+            {
+                //calculamos la rotación para pasar a un sistema que contenga a
+                //los dos primeros puntos en el ecuador, con el primero en
+                //(lat=0,lon=0)
+                RotaArco00Ecuador(tolAng,
+                                  yF[pos[0]],xF[pos[0]],yF[pos[1]],xF[pos[1]],
+                                  mRot,NULL);
+                //calculamos la latitud del tercer punto
+                AplicaMatrizRotacionCoorGeod(1,yF[pos[2]],xF[pos[2]],mRot,
+                                             &latCR,NULL);
+                //comprobamos si la latitud es 0.0
+                if(GEOC_ES_CERO(latCR,GEOC_ARC_RES_ANG))
+                {
+                    //los puntos son colineales
+                    colin = 1;
+                }
+            }
+            else
+            {
+                //comprobamos si los tres puntos son colineales
+                colin = TresPuntosColineales2D(poli->x[posIni+pos[0]],
+                                               poli->y[posIni+pos[0]],
+                                               poli->x[posIni+pos[1]],
+                                               poli->y[posIni+pos[1]],
+                                               poli->x[posIni+pos[2]],
+                                               poli->y[posIni+pos[2]]);
+            }
         }
         //comprobamos si después del aligerado queda algún polígono
         if((nPtos>3)||((nPtos==3)&&(!colin)))
@@ -1497,6 +1696,11 @@ int AligeraPolig(polig* poli,
                 free(pos);
                 free(x);
                 free(y);
+                if(facCoor!=1.0)
+                {
+                    free(xF);
+                    free(yF);
+                }
                 //mensaje de error
                 GEOC_ERROR("Error de asignación de memoria");
                 //salimos de la función
@@ -1520,6 +1724,11 @@ int AligeraPolig(polig* poli,
                 free(pos);
                 free(x);
                 free(y);
+                if(facCoor!=1.0)
+                {
+                    free(xF);
+                    free(yF);
+                }
                 //escribimos el mensaje de error
                 GEOC_ERROR("Error de asignación de memoria");
                 //salimos de la función
@@ -1544,6 +1753,11 @@ int AligeraPolig(polig* poli,
         {
             //liberamos la memoria asignada
             LibMemPolig(aux);
+            if(facCoor!=1.0)
+            {
+                free(xF);
+                free(yF);
+            }
             //mensaje de error
             GEOC_ERROR("Error de asignación de memoria");
             //salimos de la función
@@ -1554,7 +1768,7 @@ int AligeraPolig(polig* poli,
     if(poli->hayArea)
     {
         //calculamos las áreas
-        estado = CalcAreaPolig(aux);
+        estado = CalcAreaPolig(aux,facCoor,esf,a,f);
         //comprobamos los posibles errores
         if(estado!=GEOC_ERR_NO_ERROR)
         {
@@ -1571,9 +1785,15 @@ int AligeraPolig(polig* poli,
     //enlazamos los campos de la estructura auxiliar a los de la estructura de
     //salida
     EnlazaCamposPolig(aux,poli);
-    //liberamos la memoria asignada a la estructura auxiliar (pero no a sus
-    //campos)
+    ////////////////////////////////////////////////////////////////////////////
+    ////////////////////////////////////////////////////////////////////////////
+    //liberamos la memoria asignada
     free(aux);
+    if(facCoor!=1.0)
+    {
+        free(xF);
+        free(yF);
+    }
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
     //salimos de la función
@@ -1588,6 +1808,8 @@ void ImprimeCabeceraPoligFichero(const polig* poli,
                                  const char formCoor[],
                                  const int impArea,
                                  const char formArea[],
+                                 const int impAtr,
+                                 const char formAtr[],
                                  const double factorX,
                                  const double factorY,
                                  const int repitePrimerPunto,
@@ -1713,6 +1935,15 @@ shared(poli,pos,nVert,posXMin,posXMax,posYMin,posYMax)
     }
     ////////////////////////////////////////////////////////////////////////////
     ////////////////////////////////////////////////////////////////////////////
+    //comprobamos si hay que imprimir los atributos
+    if(impAtr)
+    {
+        //imprimimos el valor del atributo
+        fprintf(idFich," ");
+        fprintf(idFich,formAtr,poli->atr[indice]);
+    }
+    ////////////////////////////////////////////////////////////////////////////
+    ////////////////////////////////////////////////////////////////////////////
     //salto de línea final
     fprintf(idFich,"\n");
     ////////////////////////////////////////////////////////////////////////////
@@ -1734,6 +1965,8 @@ void ImprimePoligFichero(const polig* poli,
                          const int impLim,
                          const int impArea,
                          const char formArea[],
+                         const int impAtr,
+                         const char formAtr[],
                          FILE* idFich)
 {
     //índices para recorrer bucles
@@ -1773,7 +2006,7 @@ void ImprimePoligFichero(const polig* poli,
         {
             //imprimos la cabecera
             ImprimeCabeceraPoligFichero(poli,i,iniCab,impLim,formCoor,impArea,
-                                        formArea,factorX,factorY,
+                                        formArea,impAtr,formAtr,factorX,factorY,
                                         repitePrimerPunto,idFich);
         }
         //posición del punto inicial del polígono
